
import re
import os
import shutil
import uuid
import gzip
import logging
import sys
import glob
import itertools
from contextlib import contextmanager
from subprocess import check_output, CalledProcessError


def window(it):
    """
    >>> list(window(range(5)))
    [(None, 0, 1), (0, 1, 2), (1, 2, 3), (2, 3, 4), (3, 4, None)]
    """
    it = itertools.chain([None], it, [None])  # pad for completeness
    result = tuple(itertools.islice(it, 3))

    if len(result) == 3:
        yield result

    for elem in it:
        result = result[1:] + (elem,)
        yield result


def chunks(it, size):
    """
    Chunk a generator into a given size (last chunk might be smaller)
    """
    buf = []
    for s in it:
        buf.append(s)
        if len(buf) == size:
            yield buf
            buf = []
    if len(buf) > 0:
        yield buf


def flatten(it):
    """
    >>> list(flatten([['abc', 'cde'], ['yte']]))
    ['a', 'b', 'c', 'c', 'd', 'e', 'y', 't', 'e']
    """
    if isinstance(it, str):
        for i in it:
            yield i
    else:
        for subit in it:
            yield from flatten(subit)


def ensure_ext(path, ext, infix=None):
    """
    Compute target path with eventual infix and extension

    >>> ensure_ext("model.pt", "pt", infix="0.87")
    'model-0.87.pt'
    >>> ensure_ext("model.test", "pt", infix="0.87")
    'model-0.87.test.pt'
    >>> ensure_ext("model.test", "test", infix="pie")
    'model-pie.test'
    """
    path, oldext = os.path.splitext(path)

    # normalize extension
    if ext.startswith("."):
        ext = ext[1:]
    if oldext.startswith("."):
        oldext = oldext[1:]

    # infix
    if infix is not None:
        path = "-".join([path, infix])

    # add old extension if not the same as the new one
    if oldext and oldext != ext:
        path = '.'.join([path, oldext])

    return '.'.join([path, ext])


def get_filenames(input_path):
    """
    Get filenames from path expression
    """
    if os.path.isdir(input_path):
        filenames = [os.path.join(input_path, f) for f in os.listdir(input_path)
                     if not f.startswith('.')]
    elif os.path.isfile(input_path):
        filenames = [input_path]
    else:                       # unix string
        filenames = glob.glob(input_path)

    return filenames


@contextmanager
def shutup():
    with open(os.devnull, "w") as void:
        old_out = sys.stdout
        old_err = sys.stderr
        sys.stdout = void
        sys.stderr = void
        try:
            yield
        finally:
            sys.stdout = old_out
            sys.stderr = old_err


@contextmanager
def tmpfile(parent='/tmp/'):
    fid = str(uuid.uuid1())
    tmppath = os.path.join(parent, fid)
    yield tmppath
    if os.path.isdir(tmppath):
        shutil.rmtree(tmppath)
    else:
        os.remove(tmppath)


def add_gzip_to_tar(string, subpath, tar):
    with tmpfile() as tmppath:
        with gzip.GzipFile(tmppath, 'w') as f:
            f.write(string.encode())
        tar.add(tmppath, arcname=subpath)


def get_gzip_from_tar(tar, fpath):
    return gzip.open(tar.extractfile(fpath)).read().decode().strip()


class GitInfo():
    """
    Utility class to retrieve git-based info from a repository
    """
    def __init__(self, fname):
        if os.path.isfile(fname):
            self.dirname = os.path.dirname(fname)
        elif os.path.isdir(fname):
            self.dirname = fname
        else:
            # not a file
            self.dirname = None

        if not os.path.isfile(fname) and not os.path.isdir(fname):
            logging.warn("[GitInfo]: Input file doesn't exit")

        else:
            try:
                with shutup():
                    check_output(['git', '--version'], cwd=self.dirname)
            except FileNotFoundError:
                self.dirname = None
                logging.warn("[GitInfo]: Git doesn't seem to be installed")
            except CalledProcessError as e:
                self.dirname = None
                code, _ = e.args
                if code == 128:
                    logging.warn("[GitInfo]: Script not git-tracked")
                else:
                    logging.warn("[GitInfo]: Unrecognized git error")

    def run(self, cmd):
        if self.dirname is None:
            return

        return check_output(cmd, cwd=self.dirname).strip().decode('utf-8')

    def get_commit(self):
        """
        Returns current commit on file or None if an error is thrown by git
        (OSError) or if file is not under git VCS (CalledProcessError)
        """
        return self.run(["git", "describe", "--always"])

    def get_branch(self):
        """
        Returns current active branch on file or None if an error is thrown
        by git (OSError) or if file is not under git VCS (CalledProcessError)
        """
        return self.run(["git", "rev-parse", "--abbrev-ref", "HEAD"])

    def get_tag(self):
        """
        Returns current active tag
        """
        return self.run(["git", "describe", "--tags", "--abbrev=0"])


def model_spec(inp):
    """
    >>> example = 'model-pos-2018:03:05.tar'
    >>> model_spec(example)
    [('model-pos-2018:03:05.tar', [])]

    >>> example = '<model-pos-2018:03:05.tar,pos><model-pos-2018:03:05.tar,lemma>'
    >>> model_spec(example)
    [('model-pos-2018:03:05.tar', ['pos']), ('model-pos-2018:03:05.tar', ['lemma'])]
    """
    if not inp.startswith('<'):
        return [(inp, [])]

    output = []
    for string in re.findall(r'<([^>]+)>', inp):
        model_path, *tasks = string.split(',')
        output.append((model_path, tasks))

    return output




import os
import yaml
import json
from json_minify import json_minify


DEFAULTPATH = os.sep.join([os.path.dirname(__file__), 'default_settings.json'])


class Settings(dict):
    def __init__(self, *args, **kwargs):
        super(Settings, self).__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    self[k] = v

        if kwargs:
            for k, v in kwargs.iteritems():
                self[k] = v

    def __getattr__(self, attr):
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super(Settings, self).__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        super(Settings, self).__delitem__(key)
        del self.__dict__[key]


def flat_merge(s1, s2):
    """
    Merge two dictionaries in a flat way (non-recursive)

    >>> flat_merge({"a": {"b": 1}}, {"a": {"c": 2}})
    {'a': {'b': 1}}
    """
    for k in s2:
        if k not in s1:
            s1[k] = s2[k]

    return s1


def recursive_merge(s1, s2):
    """
    Recursively merge two dictionaries

    >>> recursive_merge({"a": {"b": 1}}, {"a": {"c": 2}})
    {'a': {'b': 1, 'c': 2}}
    """
    for k, v in s2.items():
        if k in s1 and isinstance(v, dict):
            if not isinstance(s1[k], dict):
                raise ValueError("Expected dictionary at key [{}]".format(k))
            s1[k] = recursive_merge(s1[k], v)
        elif k not in s1:
            s1[k] = v

    return s1


def merge_task_defaults(settings):
    for task in settings.tasks:
        task_settings = task.get("settings", {})
        task_settings["target"] = task_settings.get("target", task['name'])
        task['settings'] = task_settings
        for tkey, tval in settings.task_defaults.items():
            task[tkey] = task.get(tkey, tval)

    return settings


def load_default_settings():
    """
    Load built-in default settings
    """
    with open(DEFAULTPATH) as f:
        return merge_task_defaults(Settings(json.loads(json_minify(f.read()))))


def settings_from_file(config_path):
    """Loads and parses a parameter file.

    Parameters
    ===========
    config_path : str
        The path to the parameter file, formatted as json.

    Returns
    ===========
    settings : dict, A dictionary with the parameters
    """

    try:
        with open(config_path, 'r') as f:
            p = json.loads(json_minify(f.read()))
    except Exception as e:
        raise ValueError(
            "Couldn't read config file: %s. Exception: %s" % (config_path, str(e)))

    # add default values for missing settings:
    with open(DEFAULTPATH, 'r') as f:
        defaults = json.loads(json_minify(f.read()))

    # settings = Settings(flat_merge(p, defaults))
    settings = Settings(recursive_merge(p, defaults))

    # ultimately overwrite settings from environ vars of the form PIE_{var}
    checked = []
    for k in settings:
        env_k = 'PIE_{}'.format(k.upper())
        if env_k in os.environ:
            # transform to target type and overwrite settings
            settings[k] = type(defaults[k])(os.environ[env_k])
            checked.append(env_k)
    for env_k in os.environ:
        if env_k.startswith('PIE_') and env_k not in checked:
            raise ValueError(
                "Environment variable '{}' didn't match. Aborting!".format(env_k))

    # store the config path too:
    settings.config_path = config_path

    if settings.verbose:
        print("\n::: Loaded Config :::\n")
        print(yaml.dump(dict(settings)))

    return merge_task_defaults(settings)




import torch
import torch.nn as nn
import torch.nn.functional as F

from pie import torch_utils
from pie import initialization

from .encoder import RNNEncoder
from .lstm import CustomBiLSTM


class CNNEmbedding(nn.Module):
    """
    Character-level Embeddings with Convolutions following Kim 2014.
    """
    def __init__(self, num_embeddings, embedding_dim, padding_idx=None,
                 kernel_sizes=(5, 4, 3), out_channels=100):
        self.num_embeddings = num_embeddings
        self.embedding_dim = out_channels * len(kernel_sizes)
        self.kernel_sizes = kernel_sizes
        self.out_channels = out_channels
        super().__init__()

        self.emb = nn.Embedding(
            num_embeddings, embedding_dim, padding_idx=padding_idx)

        convs = []
        for W in kernel_sizes:
            wide_pad = (0, (W-1) // 2)
            conv = nn.Conv2d(
                1, out_channels, (embedding_dim, W), padding=wide_pad)
            convs.append(conv)
        self.convs = nn.ModuleList(convs)

        self.init()

    def init(self):
        initialization.init_embeddings(self.emb)
        for conv in self.convs:
            initialization.init_conv(conv)

    def forward(self, char, nchars, nwords):
        emb = self.emb(char)

        emb = emb.transpose(0, 1)  # (batch x seq_len x emb_dim)
        emb = emb.transpose(1, 2)  # (batch x emb_dim x seq_len)
        emb = emb.unsqueeze(1)     # (batch x 1 x emb_dim x seq_len)

        conv_outs, maxlen = [], 0
        for conv in self.convs:
            # (batch x C_o x seq_len)
            conv_outs.append(F.relu(conv(emb).squeeze(2)))
            maxlen = max(maxlen, conv_outs[-1].size(2))

        conv_outs = [F.pad(out, (0, maxlen - out.size(2))) for out in conv_outs]
        # (batch * nwords x C_o * len(kernel_sizes) x seq_len)
        conv_outs = torch.cat(conv_outs, dim=1)
        # (batch * nwords  x C_o * len(kernel_sizes) x 1)
        conv_out = F.max_pool1d(conv_outs, maxlen).squeeze(2)
        conv_out = torch_utils.pad_flat_batch(
            conv_out, nwords, maxlen=max(nwords).item())

        conv_outs = conv_outs.transpose(0, 2).transpose(1, 2)

        return conv_out, conv_outs


class RNNEmbedding(nn.Module):
    """
    Character-level Embeddings with BiRNNs.
    """
    def __init__(self, num_embeddings, embedding_dim, padding_idx=None,
                 custom_lstm=False, cell='LSTM', init_rnn='default',
                 num_layers=1, dropout=0.0):
        self.num_embeddings = num_embeddings
        self.num_layers = num_layers
        self.embedding_dim = embedding_dim * 2  # bidirectional
        super().__init__()

        self.emb = nn.Embedding(num_embeddings, embedding_dim, padding_idx=padding_idx)
        initialization.init_embeddings(self.emb)

        if custom_lstm:
            self.rnn = CustomBiLSTM(
                embedding_dim, embedding_dim, num_layers=num_layers, dropout=dropout)
        else:
            self.rnn = getattr(nn, cell)(
                embedding_dim, embedding_dim, bidirectional=True,
                num_layers=num_layers, dropout=dropout if num_layers > 1 else 0)
            initialization.init_rnn(self.rnn, scheme=init_rnn)

    def forward(self, char, nchars, nwords):
        """
        Parameters
        ===========
        char : tensor((seq_len x) batch)
        nchars : tensor(batch)
        nwords : tensor(output batch)
        """
        # embed
        char = self.emb(char)
        # rnn
        hidden = None
        _, sort = torch.sort(nchars, descending=True)
        _, unsort = sort.sort()
        char, nchars = char[:, sort], nchars[sort]
        if isinstance(self.rnn, nn.RNNBase):
            outs, emb = self.rnn(
                nn.utils.rnn.pack_padded_sequence(char, nchars), hidden)
            outs, _ = nn.utils.rnn.pad_packed_sequence(outs)
            if isinstance(emb, tuple):
                emb, _ = emb
        else:
            outs, (emb, _) = self.rnn(char, hidden, nchars)
        # (max_seq_len x batch * nwords x emb_dim)
        outs, emb = outs[:, unsort], emb[:, unsort]
        # (layers * 2 x batch x hidden) -> (layers x 2 x batch x hidden)
        emb = emb.view(self.num_layers, 2, len(nchars), -1)
        # use only last layer
        emb = emb[-1]
        # (2 x batch x hidden) - > (batch x 2 * hidden)
        emb = emb.transpose(0, 1).contiguous().view(len(nchars), -1)
        # (batch x 2 * hidden) -> (nwords x batch x 2 * hidden)
        emb = torch_utils.pad_flat_batch(emb, nwords, maxlen=max(nwords).item())

        return emb, outs


class EmbeddingMixer(nn.Module):
    def __init__(self, emb_dim):
        self.embedding_dim = emb_dim
        super().__init__()

        # mix parameter
        self.alpha = nn.Parameter(
            torch.Tensor(self.embedding_dim * 2, 1).uniform_(-0.05, 0.05))

    def forward(self, wembs, cembs):
        # ((seq_len x) batch x emb_dim * 2)
        alpha_in = torch.cat([wembs, cembs], dim=-1)
        # ((seq_len x) batch)
        if wembs.dim() == 3:
            alpha = F.sigmoid(torch.einsum('do,mbd->mb', [self.alpha, alpha_in]))
        else:
            alpha = F.sigmoid(torch.einsum('do,bd->b', [self.alpha, alpha_in]))

        wembs = alpha.unsqueeze(-1).expand_as(wembs) * wembs
        cembs = (1 - alpha).unsqueeze(-1).expand_as(cembs) * cembs

        return wembs + cembs


def EmbeddingConcat():
    def func(wemb, cemb):
        return torch.cat([wemb, cemb], dim=-1)
    return func


if __name__ == '__main__':
    from pie.settings import settings_from_file
    from pie.data import Dataset

    settings = settings_from_file('./config.json')
    data = Dataset(settings)
    ((word, wlen), (char, clen)), tasks = next(data.batch_generator())
    print("lemma", tasks['lemma'][0].size(), tasks['lemma'][1])
    print("char", char.size(), clen)
    print("word", word.size(), wlen)

    emb_dim = 20
    wemb = nn.Embedding(len(data.label_encoder.word), emb_dim)
    cemb = RNNEmbedding(len(data.label_encoder.char), emb_dim)
    cnncemb = CNNEmbedding(len(data.label_encoder.char), emb_dim)

    mixer = EmbeddingMixer(20)
    w, (c, _) = wemb(word), cemb(char, clen, wlen)
    output = mixer(w, c)

    output2 = []
    for w, c in zip(w, c):
        output2.append(mixer(w, c))
    output2 = torch.stack(output2)

    print(output.size())
